from ... domains.interface import Interface
import jinja2
import os

RELATIVE_SCRIPT_PATH = "first_logon_script.ps1"

def get_logon_script(interfaces: list[Interface], ssh_pub: str, range_default_gateway: str, range_dns_server: str, management_default_gateway: str):
     nets = [] 
     for interface in interfaces:
         interface_gw = ""
         interface_dns = ""
         if not interface.is_mngmt:
             interface_gw = range_default_gateway
             interface_dns = range_dns_server
         else:
             interface_gw = management_default_gateway
             interface_dns = management_default_gateway

         nets.append([interface.mac, interface.ipv4, interface.prefix_length, interface_gw, interface_dns, interface.is_mngmt])

     config = {
         "interfaces": nets,
         "ssh_pub": ssh_pub
     }

     script_path = f"{os.path.dirname(
         __file__)}/{RELATIVE_SCRIPT_PATH}"

     with open(script_path, "r") as f:
         template = jinja2.Template(f.read())
         cont = template.render(config)
         return cont




import libvirt
import sys
sys.path.append('../cave')

# setting path
from cave import Interface 
from cave import Range
import json, os
import subprocess
import logging

HOST = "192.168.188.26"
REMOTE_IMAGES_DIR = "/opt/test/images"
REMOTE_POOL_DIR = "/opt/test/pool1"

CON_URI = f"qemu+ssh://root@{HOST}/system"
SSH_URI = f"root@{HOST}"
TMP_CLOUD_INIT_DIR = "/tmp/cloud_init"

UBUNTU_IMAGE = "/opt/IMG/noble-server-cloudimg-amd64.qcow2"
WINDOWS_SERVER_ISO = "/opt/ISO/winserver22.iso"
LOCAL_IMAGE_PATHS = [WINDOWS_SERVER_ISO, UBUNTU_IMAGE]

USERNAME = "user"
PASSWORD = "banga"

SSH_PUB = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIKXiXbuBe8utnnR7RlgihIlJMDTB8i+92+hLl97uHw18"
SSH_PRIV_PATH = "/home/snoja/.ssh/bac/test"
LOGGER = logging.getLogger()
### setup host for test run, push image/iso if it does not exits
return_code = subprocess.call(["ssh", SSH_URI, f"mkdir -p {TMP_CLOUD_INIT_DIR} {REMOTE_IMAGES_DIR} {REMOTE_POOL_DIR}"], stdout=subprocess.DEVNULL)

for local_image_path in LOCAL_IMAGE_PATHS:
    # upload images
    file_name = os.path.basename(local_image_path)
    remote_image_path = f"{REMOTE_IMAGES_DIR}/{file_name}"
    return_code = subprocess.call(["ssh", SSH_URI, f"test -f {remote_image_path}"], stdout=subprocess.DEVNULL)
    if return_code != 0:
        return_code = subprocess.call(["scp", local_image_path, f"{SSH_URI}:{remote_image_path}"], stdout=subprocess.DEVNULL)
        assert return_code == 0, f"Upload of image {local_image_path} to {remote_image_path} failed"
    else:
        LOGGER.info(f"{file_name}")

conn = libvirt.open(CON_URI)

range = Range(conn, SSH_URI, TMP_CLOUD_INIT_DIR, SSH_PUB, USERNAME, PASSWORD)
range.nuke_libvirt()
range.add_pool("pool1",REMOTE_POOL_DIR)

# isolated network
mngt = range.add_management_network(name="mngmt", 
                         ipv4="10.10.0.1", 
                         ipv4_subnet="255.255.255.0")

n1 = range.add_network(name="network1", 
                       ipv4="", 
                       ipv4_subnet="", 
                       mode="")

i1 = Interface(mac="", 
               network=n1, 
               ipv4="10.10.1.2", 
               prefix_length=24, 
               is_mngmt=False) 

im = Interface(mac="", 
               network=mngt, 
               ipv4="10.10.0.2", 
               prefix_length=24, 
               is_mngmt=True) 

ubuntu_img = f"{REMOTE_IMAGES_DIR}/{os.path.basename(UBUNTU_IMAGE)}"
ld = range.add_linux_domain(name="linux01", 
                            hostname="lin01", 
                            base_image_path=ubuntu_img, 
                            interfaces=[i1, im],
                            graphics_passwd="pw",
                            disk_volume_size_gb=8,
                            memory=1024,
                            vcpus=2,
                            graphics_port="6000",
                            graphics_auto_port="no", 
                            graphics_address="0.0.0.0",
                            default_gateway="10.10.1.1",
                            dns_server="1.1.1.1",
                            management_default_gateway="10.10.0.1") 

i1 = Interface(mac="", 
               network=n1, 
               ipv4="10.10.1.4", 
               prefix_length=24, 
               is_mngmt=False) 

im = Interface(mac="", 
               network=mngt, 
               ipv4="10.10.0.4", 
               prefix_length=24, 
               is_mngmt=True) 

ld = range.add_linux_domain("linux02", "lin02", ubuntu_img, [i1, im], "pw", 8, 1024, 2, "6001", "no", "0.0.0.0", "10.10.1.1", "1.1.1.1", "10.10.0.1") 

i1 = Interface(mac="", 
               network=n1, 
               ipv4="10.10.1.3", 
               prefix_length=24, 
               is_mngmt=False) 

im = Interface(mac="", 
               network=mngt, 
               ipv4="10.10.0.3",
               prefix_length=24, 
               is_mngmt=True) 

windows_server22_iso = f"{REMOTE_IMAGES_DIR}/{os.path.basename(WINDOWS_SERVER_ISO)}"
wd = range.add_windows_server_domain("windows01", "win01", windows_server22_iso, None, 2, [i1,im],"pw", 20, 3000, 2, "6002", "no", "0.0.0.0", "10.10.1.1", "1.1.1.1", "10.10.0.1" )

range.block_unil_rdy()

# TODO: cleanup

# testcase whoami
LOGGER.info("testcase whoami")

# windows01
return_val = subprocess.check_output(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.3", f'PowerShell -Command "echo $Env:UserName"'])
assert return_val.decode().startswith(USERNAME)

# linux01
return_val = subprocess.check_output(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.2", f'whoami'])
assert return_val.decode().startswith(USERNAME)

# linux02
return_val = subprocess.check_output(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.4", f'whoami'])
assert return_val.decode().startswith(USERNAME)

# network/ping tests
# windows01
return_val = subprocess.call(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.3", f'PowerShell -Command "(ping 10.10.1.2) -and (ping 10.10.1.4)"'])
assert return_val == 0

# linux01
return_val = subprocess.call(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.2", f'ping 10.10.1.3 -c 2 -4 && ping 10.10.1.4 -c 2 -4'])
assert return_val == 0

# linux02
return_val = subprocess.call(["ssh", "-i", SSH_PRIV_PATH, "-o", "UserKnownHostsFile /dev/null", "-o StrictHostKeyChecking=no", f"{USERNAME}@10.10.0.4", f'ping 10.10.1.2 -c 2 -4 && ping 10.10.1.3 -c 2 -4'])
assert return_val == 0

range.cleanup()



from ..networks.network.network import Network
from dataclasses import dataclass
@dataclass
class Interface:
    mac: str
    network: Network
    ipv4: str
    prefix_length: int
    is_mngmt: bool




from ..templated_libvirt_domain import TemplatedLibvirtDomain
from ..interface import Interface
import os, uuid
from ...os_types import OsType

class WindowsDomain(TemplatedLibvirtDomain):
    RELATIVE_TEMPLATE_PATH = "domain_windows.jinja.xml"
    def get_template_str(self):
        template_path = f"{os.path.dirname(
            __file__)}/{WindowsDomain.RELATIVE_TEMPLATE_PATH}"

        with open(template_path, "r") as f:
            template_content = f.read()

        return template_content
            
    def __init__(self, name: str, disk_volume, boot_iso, interfaces: list[Interface], graphics_passwd: str, cdrom: str, memory: int, vcpus: int, graphics_port: str, graphics_auto_port: str, graphics_address: str, os_type:OsType):
        assert not (graphics_port and (graphics_auto_port == "yes"))
        config = {
            "name": name,
            "memory": memory,
            "uuid": uuid.uuid4(),
            "vcpus": vcpus,
            "cdrom": cdrom,
            "boot_iso": boot_iso,
            "interfaces": [{"mac": x.mac, "net_name" : x.network.name, "net_host_mac": x.network.host_mac, "net_host_isolated": x.network.host_isolated} for x in interfaces],
            "disk_volume": {
                "format": disk_volume.format,
                "name": disk_volume.name,
                "base_img": disk_volume.base_img,
                "pool_name": disk_volume.pool_name
            },
            "graphics_port": graphics_port,
            "graphics_passwd": graphics_passwd,
            "graphics_auto_port": graphics_auto_port,
            "graphics_address": graphics_address
        }

        super().__init__(config, self.get_template_str(), name, interfaces, os_type)

        self.name = name
        self.disk_volume = disk_volume
        self.interfaces = interfaces
        self.graphics_passwd = graphics_passwd
        self.cdrom = cdrom
        self.boot_iso = boot_iso
        self.memory = memory
        self.vcpus = vcpus
        self.graphics_port = graphics_port
        self.graphics_auto_port = graphics_auto_port
        self.graphics_address = graphics_address 
        self.os_type = os_type


    def create(self, conn):
        super().create(conn)

    @staticmethod
    def remove_cdrom(conn, name):
        libvirt_domain = conn.lookupByName(name)
        # TODO: do proper xml parsing, don not hardcode ua-config-cdrom
        target = f"""
            <disk type='file' device='cdrom'>
                <driver name='qemu' type='raw'/>
                <backingStore/>
                <target dev='hdc' bus='sata'/>
                <readonly/>
                <alias name='ua-config-cdrom'/>
            </disk>
        """
        libvirt_domain.updateDeviceFlags(target, 0x3)




import jinja2
import os
import uuid
from ...domains.interface import Interface

class CloudInitUserConfig():
    RELATIVE_TEMPLATE_PATH = "cloud-init-user.jinja.yaml"
    CLOUDINIT_FILE_NAME = "user-data"

    def __init__(self, username: str, password: str, ssh_pub: str, hostname="ab"):
        self.username = username
        self.password = password
        self.ssh_pub = ssh_pub
        self.hostname = hostname

    def _get_config(self):
        config = {
            "username": self.username,
            "password": self.password,
            "ssh_pub": self.ssh_pub,
            "hostname": self.hostname
        }
        return config 

    def __repr__(self):
        template_path = f"{os.path.dirname(
            __file__)}/{CloudInitUserConfig.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            return template.render(self._get_config())

class CloudInitMetaData():
    RELATIVE_TEMPLATE_PATH = "cloud-init-meta.jinja.yaml"
    CLOUDINIT_FILE_NAME = "meta-data"

    def __init__(self,  hostname: str, instance_id=""):
        self.instance_id = instance_id if instance_id else str(uuid.uuid4())
        self.hostname = hostname

    def _get_config(self):
        config = {
            "instance_id": self.instance_id,
            "local_hostname": self.hostname
        }
        return config

    def __repr__(self):
        template_path = f"{os.path.dirname(
            __file__)}/{CloudInitMetaData.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            return template.render(self._get_config())

class CloudInitNetworkConfig():
    RELATIVE_TEMPLATE_PATH = "cloud-init-network.jinja.yml"
    CLOUDINIT_FILE_NAME = "network-config"

    def __init__(self,  interfaces: list[Interface], range_default_gateway:str, range_dns_server: str, management_default_gateway: str):
        self.interfaces = interfaces
        self.management_default_gateway = management_default_gateway
        self.range_default_gateway = range_default_gateway
        self.range_dns_server = range_dns_server

    def _get_config(self):
        config = {
            "interfaces": [[interface.mac, f"{interface.ipv4}/{interface.prefix_length}", interface.is_mngmt] for interface in self.interfaces],
            "management_default_gateway": self.management_default_gateway,
            "range_default_gateway": self.range_default_gateway,
            "range_dns_server": self.range_dns_server
        }
        return config

    def __repr__(self):
        template_path = f"{os.path.dirname(
            __file__)}/{CloudInitNetworkConfig.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            cont = template.render(self._get_config())
            return cont



from libvirt import libvirtError
from .src import Pool
from .src import Network
from .src import OsType
from .src import LinuxDomain
from .src import Interface
from .src import CloudInitMetaData, CloudInitUserConfig, CloudInitNetworkConfig
from .src import create_and_push_cd
from .src import Autounattend
from .src import WindowsDomain
from .src import WindowsDomain
from .src import AutounattendServer
from .src import WindowsTPMDomain
from .src import AutounattendGPT

import uuid
import logging
import socket
from time import sleep
from random import randint
import sys

root = logging.getLogger("default")
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
logger = logging.getLogger("default")

def random_mac():
    return "02:00:00:%02x:%02x:%02x" % (randint(0, 255),
                                        randint(0, 255),
                                        randint(0, 255))

class Range():
    def __init__(self, libvirt_connection, ssh_connect_uri: str, remote_autoinstall_iso_path: str, domain_ssh_pub_key: str, domain_user: str, domain_password: str):
        self.libvirt_connection = libvirt_connection
        self.ssh_connect_uri = ssh_connect_uri
        self.remote_autoinstall_iso_path = remote_autoinstall_iso_path
        self.domain_user = domain_user
        self.domain_password = domain_password
        self.domain_ssh_pub_key = domain_ssh_pub_key
        self.pool = None
        self.networks = []
        self.domains = []
        self.management_network = None

    def add_pool(self, name: str, path: str):
        self.pool = Pool(name, path)
        self.pool.create(self.libvirt_connection)
        return self.pool
         
    def add_network(self, name: str, mode: str, ipv4: str, ipv4_subnet: str, ingress_route_subnet=None, ingress_route_gateway=None):

        n = Network(name=name, 
                    host_isolated=True if mode == "" else False, 
                    ipv4=ipv4, ipv4_subnet=ipv4_subnet, 
                    isolate_guests=False,
                    ipv6="", ipv6_prefix="", 
                    mode=mode, 
                    ingress_route_subnet=ingress_route_subnet, ingress_route_gateway=ingress_route_gateway)


        n.create(self.libvirt_connection)
        self.networks.append(n)
        return n

    def add_management_network(self, name: str, ipv4: str, ipv4_subnet: str):
        if self.management_network:
            logger.error("Multiple management networks for a single range is not supported for now")
            return

        n = Network(name=name, 
                    host_isolated=False, 
                    ipv4=ipv4, ipv4_subnet=ipv4_subnet, 
                    isolate_guests=False,
                    ipv6="", ipv6_prefix="", 
                    mode="open", 
                    ingress_route_subnet=None, ingress_route_gateway=None)

        n.create(self.libvirt_connection)
        self.networks.append(n)
        self.management_network = n
        return n



    def add_linux_domain(self, name: str, hostname: str, base_image_path: str, interfaces: list[Interface], graphics_passwd: str,  disk_volume_size_gb: int, memory: int, vcpus: int, graphics_port: str, graphics_auto_port: str, graphics_address: str, default_gateway: str, dns_server: str, management_default_gateway: str):
        """
            Args:
        """
        if not self.pool:
            logger.error(f"cannot create linux domain before pool was created")
        assert self.pool

        # generate random macs for interfaces without mac
        for interface in interfaces:
            if len(interface.mac) == 0:
                interface.mac = random_mac()
        
        meta_data = CloudInitMetaData(hostname=hostname, instance_id=str(uuid.uuid4())) 
        user_config = CloudInitUserConfig(self.domain_user, 
                        self.domain_password,
                        self.domain_ssh_pub_key 
                        )
        network_config = CloudInitNetworkConfig(interfaces, 
                            range_default_gateway=default_gateway,
                            range_dns_server=dns_server, 
                            management_default_gateway=management_default_gateway)
        
        cdrom_files = {CloudInitMetaData.CLOUDINIT_FILE_NAME:str(meta_data), 
                        CloudInitUserConfig.CLOUDINIT_FILE_NAME:str(user_config),
                        CloudInitNetworkConfig.CLOUDINIT_FILE_NAME:str(network_config)}

        remote_path = create_and_push_cd(f"{self.remote_autoinstall_iso_path}/{uuid.uuid4()}.iso", self.ssh_connect_uri, cdrom_files)

        volume = self.pool.create_volume(f"{name}_vol", disk_volume_size_gb, base_image_path, format="qcow2")
        linux_domain = LinuxDomain(name, volume, interfaces, graphics_passwd, remote_path, memory, vcpus, graphics_port, graphics_auto_port, graphics_address, OsType.GENERIC_LINUX)
        linux_domain.create(self.libvirt_connection)
        self.domains.append(linux_domain)
        return linux_domain

    def add_windows_domain(self, name: str, hostname: str, boot_iso: str, product_key: str, interfaces: list[Interface], graphics_passwd: str,  disk_volume_size_gb: int, memory: int, vcpus: int, graphics_port: str, graphics_auto_port: str, graphics_address: str, default_gateway: str, dns_server: str, management_default_gateway: str):
        if not self.pool:
            logger.error(f"cannot create windows domain before pool was created")

        assert self.pool

        # generate random macs for interfaces without mac
        for interface in interfaces:
            if len(interface.mac) == 0:
                interface.mac = random_mac()
        
        
        unattend = Autounattend(interfaces=interfaces, 
                        username=self.domain_user,
                        ssh_pub=self.domain_ssh_pub_key,
                        password=self.domain_password,
                        hostname=hostname,
                        management_default_gateway=management_default_gateway,
                        range_default_gateway=default_gateway,
                        range_dns_server=dns_server,
                        product_key=product_key
                        )

        cdrom_files = {"Autounattend.xml":str(unattend)}
        remote_path = create_and_push_cd(f"{self.remote_autoinstall_iso_path}/{uuid.uuid4()}.iso", self.ssh_connect_uri, cdrom_files)

        volume = self.pool.create_volume(f"{name}_vol", disk_volume_size_gb, None, format="raw")

        wd = WindowsDomain(name, volume, boot_iso, interfaces, graphics_passwd, remote_path, memory, vcpus, graphics_port, graphics_auto_port, graphics_address, OsType.GENERIC_WINDOWS)

        wd.create(self.libvirt_connection)
        self.domains.append(wd)
        return wd

    def add_windows_server_domain(self, name: str, hostname: str, boot_iso: str, product_key: str|None, version_index: int|None, interfaces: list[Interface], graphics_passwd: str,  disk_volume_size_gb: int, memory: int, vcpus: int, graphics_port: str, graphics_auto_port: str, graphics_address: str, default_gateway: str, dns_server: str, management_default_gateway: str):
        if not self.pool:
            logger.error(f"cannot create windows server domain before pool was created")

        assert self.pool

        # generate random macs for interfaces without mac
        for interface in interfaces:
            if len(interface.mac) == 0:
                interface.mac = random_mac()
        
        
        unattend = AutounattendServer(interfaces=interfaces, 
                        username=self.domain_user,
                        ssh_pub=self.domain_ssh_pub_key,
                        password=self.domain_password,
                        hostname=hostname,
                        management_default_gateway=management_default_gateway,
                        range_default_gateway=default_gateway,
                        range_dns_server=dns_server,
                        version_index=version_index,
                        product_key=None,
                        )

        cdrom_files = {"Autounattend.xml":str(unattend)}
        remote_path = create_and_push_cd(f"{self.remote_autoinstall_iso_path}/{uuid.uuid4()}.iso", self.ssh_connect_uri, cdrom_files)

        volume = self.pool.create_volume(f"{name}_vol", disk_volume_size_gb, None, format="raw")

        wd = WindowsDomain(name, volume, boot_iso, interfaces, graphics_passwd, remote_path, memory, vcpus, graphics_port, graphics_auto_port, graphics_address, OsType.GENERIC_WINDOWS_SERVER)

        wd.create(self.libvirt_connection)
        self.domains.append(wd)
        return wd

    def add_windows_tpm_domain(self, name: str, hostname: str, boot_iso: str, product_key: str, interfaces: list[Interface], graphics_passwd: str,  disk_volume_size_gb: int, memory: int, vcpus: int, graphics_port: str, graphics_auto_port: str, graphics_address: str, default_gateway: str, dns_server: str, management_default_gateway: str):
        if not self.pool:
            logger.error(f"cannot create windows domain before pool was created")

        assert self.pool

        # generate random macs for interfaces without mac
        for interface in interfaces:
            if len(interface.mac) == 0:
                interface.mac = random_mac()
        
        
        unattend = AutounattendGPT(interfaces=interfaces, 
                        username=self.domain_user,
                        ssh_pub=self.domain_ssh_pub_key,
                        password=self.domain_password,
                        hostname=hostname,
                        management_default_gateway=management_default_gateway,
                        range_default_gateway=default_gateway,
                        range_dns_server=dns_server,
                        product_key=product_key
                        )

        cdrom_files = {"Autounattend.xml":str(unattend)}
        remote_path = create_and_push_cd(f"{self.remote_autoinstall_iso_path}/{uuid.uuid4()}.iso", self.ssh_connect_uri, cdrom_files)

        volume = self.pool.create_volume(f"{name}_vol", disk_volume_size_gb, None, format="raw")

        wd = WindowsTPMDomain(name, volume, boot_iso, interfaces, graphics_passwd, remote_path, memory, vcpus, graphics_port, graphics_auto_port, graphics_address, OsType.WINDOWS_TPM)

        wd.create(self.libvirt_connection)
        self.domains.append(wd)
        return wd

    def nuke_libvirt(self):
        # Get list of all pool names (active and inactive)
        logger.info(f"nuking pools and volumes")
        pools = self.libvirt_connection.listAllStoragePools()
        for pool in pools:
            try:
                logger.info(f"Processing pool: {pool.name()}")

                # Delete all volumes in the pool
                volumes = pool.listAllVolumes()
                for volume in volumes:
                    logger.info(f"Deleting volume: {volume.name()}")
                    volume.delete(0)

                # Destroy the pool if active
                if pool.isActive():
                    logger.info(f"Destroying pool: {pool.name()}")
                    pool.destroy()

                # Undefine the pool (removes it from libvirt config)
                logger.info(f"Undefining pool: {pool.name()}")
                pool.undefine()

            except libvirtError as e:
                logger.error(f"Error processing pool '{pool.name()}': {e}")
        logger.info("All pools and volumes destroyed.")

        logger.info("Nuking domains")
        domains = self.libvirt_connection.listAllDomains()

        for domain in domains:
            try:
                if domain.isActive():
                    domain.destroy()
                if domain.isPersistent():
                    domain.undefine()
            except libvirtError as e:
                logger.error(f"Error processing domain '{domain.name()}': {e}")

        logger.info("Nuking networks")
        networks = self.libvirt_connection.listAllNetworks()

        for network in networks:
            try:
                if network.isActive():
                    network.destroy()
                if network.isPersistent():
                    network.undefine()
            except libvirtError as e:
                logger.error(f"Error processing network '{network.name()}': {e}")

    def get_state(self):
        assert self.pool
        return {"domains":[domain.to_dict() for domain in self.domains], "networks": [network.to_dict() for network in self.networks], "pool": {"name":self.pool.name, "path":self.pool.path}}

    def cleanup(self):
        logger.info(f"removing management network {self.management_network.name} {self.management_network.libvirt_network.name()}")
        for domain in self.domains:
            logger.info(f"removing management network interface for domain {domain.name}")
            domain.remove_interface_for_network_name(self.management_network.name)
        self.management_network.rm()
        self.networks.remove(self.management_network)
        logger.info(f"done removing management network {self.management_network.name}")

    def block_unil_rdy(self):
        for domain in self.domains:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            ipv4 = domain.get_mngmt_ipv4()
            result = 1
            logger.info(f"waiting for {domain.name} to be reachable")
            while result != 0:
                result = sock.connect_ex((ipv4, 22))
                sleep(5)
            logger.info(f"{domain.name} is up.")
            sock.close()




import jinja2
from logging import getLogger
from ..os_types import OsType
import os
from libvirt import libvirtError
from .interface import Interface  

logger = getLogger("default")
class TemplatedLibvirtDomain():
    RELATIVE_FILTER_PATH = "filter.xml"
    FILTER_NAME = "deny-vm-to-host"

    def __init__(self, config: dict, template_str: str, name: str, interfaces: list[Interface], os_type: OsType):
        self.config = config
        self.jinja_template = jinja2.Template(template_str) 
        self.os_type = os_type
        self.libvirt_domain = None
        self.name = name
        self.interfaces = interfaces

    def define_filter_if_not_exists(self, conn):
        filter_path = f"{os.path.dirname(
            __file__)}/{TemplatedLibvirtDomain.RELATIVE_FILTER_PATH}"
        with open(filter_path, "r") as f:
            filter = f.read()
        try:
            conn.nwfilterDefineXML(filter)
        except libvirtError:
            if str(libvirtError) == "virNWFilterDefineXML() failed":
                pass
        
    def create(self, conn):
        self.define_filter_if_not_exists(conn)
        logger.debug(f"creating libvirt domain with config: {self.config} {self.jinja_template}")
        xml_content = self.jinja_template.render(self.config)
        self.libvirt_domain = conn.defineXML(xml_content)
        self.libvirt_domain.create()

    def to_dict(self):
        dict_interfaces = []
        for interface in self.interfaces:
            dict_interfaces.append({
                "mac": interface.mac,
                "network_name": interface.network.name,
                "ipv4": interface.ipv4,
                "is_mngmt": interface.is_mngmt
            })

        return {"name": self.name, "interfaces": dict_interfaces, "os_type":str(self.os_type)}
        
    def get_mngmt_ipv4(self):
        if len(self.interfaces) < 1:
            return ""
        return next(filter(lambda x: x.is_mngmt, self.interfaces)).ipv4

    def remove_interface_for_network_name(self, network_name: str):
        assert self.libvirt_domain != None, "libvirt domain is none cannot remove interface"
        self.libvirt_domain.detachDeviceAlias(f"ua-{network_name}", 0x03)





import os
import jinja2

class Volume():
    """
        Libvirt volume
        backing image (base_img) and volume only support qcow2 format for now.
        Args:
            pool: the pool in which the vol should be created
            name: name of the volume
            path: host path of the image file
            capacity: the initial size of the volume on the host
            base_img: name of qcow2 image to use as base, this is copy on write
    """
    RELATIVE_TEMPLATE_PATH = "volume.jinja.xml"

    def __init__(self, name: str, path: str, capacity, pool_name:str, base_img=None, format="qcow2"):
        self.name = name
        self.path = path
        self.capacity = capacity
        self.base_img = base_img
        self.format = format
        self.pool_name = pool_name

    def _get_config(self) -> dict:
        config = {
            "name": self.name,
            "capacity": self.capacity,
            "path": self.path,
            "base_img": self.base_img,
            "format": self.format
        }
        return config

    def create(self, libvirt_pool):
        template_path = f"{os.path.dirname(
            __file__)}/{Volume.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            libvirt_pool.createXML(
                template.render(self._get_config()))



import os
import jinja2
from .volume import Volume


class Pool():
    """
        libvirt Pool (only dir-pools supported for now)
        Args:
            name: name of the pool
            path: the host path to the directory, directory has to exist
    """
    RELATIVE_TEMPLATE_PATH = "pool.jinja.xml"

    def __init__(self, name: str, path: str, libvirt_pool=None):
        self.name = name
        self.path = path
        self.libvirt_pool = libvirt_pool

    def _get_config(self):
        config = {
            "name": self.name,
            "path": self.path,
        }

        return config

    def create(self, conn):
        template_path = f"{os.path.dirname(
            __file__)}/{Pool.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            self.libvirt_pool = conn.storagePoolDefineXML(
                template.render(self._get_config())
            )
            self.libvirt_pool.create()

    def create_volume(self, name: str, capacity, base_img=None, format="qcow2"):
        """
            Create a volume inside this pool
            Args:
                name: the name of the volume
                capacity: the size of the volume in GiB
                base_img: None or path to the qcow2 base image.
                format: format of the volume, if not qcow2 base_image cannot be used.
        """
        if not self.libvirt_pool:
            raise Exception("cannot create volume from pool that has not been created yet")
        volume_path = f"{self.path}/{name}"
        vol = Volume(name, volume_path,
                     capacity=capacity, pool_name=self.name, base_img=base_img, format=format)
        vol.create(self.libvirt_pool)
        return vol

    @staticmethod
    def destroy_by_name(conn, name):
        pool = conn.storagePoolLookupByName(name)
        for vol in pool.listAllVolumes():
            vol.delete()
        pool.destroy()
        pool.undefine()



from .pools.dir_pool.pool import Pool
from .networks.network.network import Network
from .os_types import OsType
from .domains.domain_linux.domain_linux import LinuxDomain
from .domains.interface import Interface

from .domains.domain_windows.domain_windows import WindowsDomain

from .domains.domain_windows_tpm.domain_windows_tpm import WindowsTPMDomain

from .autoinstall.linux.cloudinit import CloudInitMetaData, CloudInitUserConfig, CloudInitNetworkConfig
from .autoinstall.cdrom import create_and_push_cd
from .autoinstall.windows.windows_client.autounattend import Autounattend
from .autoinstall.windows.windows_server.autounattend import AutounattendServer
from .autoinstall.windows.windows_gpt_tpm.autounattend import AutounattendGPT




from enum import Enum
class OsType(Enum):
    GENERIC_WINDOWS = "GENERIC_WINDOWS"
    GENERIC_WINDOWS_SERVER = "GENERIC_WINDOWS_SERVER"
    GENERIC_LINUX = "GENERIC_LINUX"
    WINDOWS_TPM = "WINDOWS_TPM"



import subprocess
import tempfile
import os
import uuid
import logging

logger = logging.getLogger("default")

def send_file(ssh_uri: str, local_file_path: str, remote_file_path: str):
    """
        Creates target folder if it does not exist and moves file_path in it
    """
    subprocess.check_call(
        ["ssh", f"{ssh_uri}", "mkdir", "-p", os.path.dirname(remote_file_path)], stdout=subprocess.DEVNULL)
    subprocess.check_call(["scp", local_file_path, f"{ssh_uri}:{remote_file_path}"], stdout=subprocess.DEVNULL)

def create_and_push_cd(remote_path, ssh_connect_string, cdrom_files: dict, ):
    logger.debug(f"Creating iso file from files: {cdrom_files}")
    with Cdrom(files=cdrom_files) as cd:
        iso = cd
        logger.debug(f"sending iso file {cd}, to {remote_path} via {ssh_connect_string}") 
        send_file(ssh_connect_string, iso, remote_path)
    return remote_path

class Cdrom():
    """
        Class used to create iso disk image from files
        Context manager (with) is supported
        Args:
            files: a dict in the format {filename: file_content, filename2: file_content2}
        Example usage:
            with Cdrom({"mydisk.iso":"Hello World!"}) as tmp_file_path:
                print(tmp_file_path)

    """
    def __init__(self, files: dict[str,str]):
        self.files = files
    def _to_iso(self, iso_filename):
        subprocess.check_call(["genisoimage", "-output", os.path.join(self.dir.name, iso_filename), "-volid", "cidata", "-joliet", "-rock", "-input-charset", "utf-8", self.dir.name], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        
    def __enter__(self):
        self.dir = tempfile.TemporaryDirectory()

        for filename,content in self.files.items():
            with open(os.path.join(self.dir.name, filename), "w", newline="\r\n") as f:
                f.write(content)

        iso_filename = f"{uuid.uuid4()}.iso"
        self._to_iso(iso_filename)

        return os.path.join(self.dir.name, iso_filename)

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.dir.cleanup()





from setuptools import setup, find_packages

project_folder = "cave"
with open('README.md') as f:
    long_desc = f.read()

setup(
    name='cave-infra',
    version='0.1.4',
    packages=find_packages(),
    include_package_data=True,
    license_files=['LICENSE.txt'],
    description='Automation toolkit for automated provisioning virtual infrastructure.',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    install_requires=[
        "Jinja2==3.1.6",
        "libvirt-python==11.4.0",
        "MarkupSafe==3.0.2"
    ],
    url='https://github.com/sn0ja/cave/',
    author='sn0ja',
)



import jinja2
import os
from ..logon_script_generator import get_logon_script
import logging

logger = logging.getLogger("default")

class AutounattendServer():
    """
        Args:
            Interfaces: [[mac, ipv4, prefix_length]] used to assign static ip to interfaces differentiating them by mac
    """
    RELATIVE_TEMPLATE_PATH = "./autounattend_windows_server.jinja.xml"
    RELATIVE_SCRIPT_PATH = "../first_logon_script.ps1"
    def __init__(self, username: str, ssh_pub: str, password, hostname: str, interfaces, management_default_gateway: str, range_default_gateway: str, range_dns_server: str, product_key: str|None, version_index: int|None):
        self.username = username
        self.ssh_pub = ssh_pub
        self.hostname = hostname
        self.v4release = True
        self.v6release = True
        self.password = password
        self.product_key = product_key
        self.interfaces = interfaces
        self.management_default_gateway = management_default_gateway
        self.range_default_gateway = range_default_gateway
        self.range_dns_server = range_dns_server
        self.version_index = version_index
        if not version_index and not product_key:
            logger.warning(f"neither product_key nor version_index specified in windows_server autounattend for hostname {hostname}")

        if product_key and version_index:
            logger.warning(f"both product_key and version_index specified in windows_server autounattend for hostname {hostname}")

    def _get_config(self):
        config = {
            "username": self.username,
            "hostname": self.hostname,
            "v4release": self.v4release,
            "v6release": self.v6release,
            "password": self.password,
            "product_key": self.product_key,
            "interfaces": self.interfaces,
            "version_index": self.version_index,
            "first_logon_script": get_logon_script(self.interfaces, 
                                                   self.ssh_pub, 
                                                   self.range_default_gateway, 
                                                   self.range_dns_server, 
                                                   self.management_default_gateway)
        }
        return config

    def __repr__(self):
        template_path = f"{os.path.dirname(
            __file__)}/{AutounattendServer.RELATIVE_TEMPLATE_PATH}"

        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            cont = template.render(self._get_config())
            return cont



import os 
import ipaddress
import jinja2
import re

class Network(object):
    """
        
        Args:
            conn: libvirt connection object
            mode: set this to empty string for isolated virt-network
            ipv4: first ip in the network, not !net-id!, if dhcp or dns is used this will be the address of the dhcp/dns server
            ipv4_subnetmask: subnetmask e.g. 255.255.255.0
            ipv6: first ipv6 in ipv6 range of network
            ipv6_prefix: e.g. 64

    """
    RELATIVE_TEMPLATE_PATH = "network.jinja.xml"

    def __init__(self, name: str,  host_isolated=None, ipv4="", ipv4_subnet="", isolate_guests=None, ipv6="", ipv6_prefix="64", mode="", ingress_route_subnet=None, ingress_route_gateway=None ):
        self.name = name
        self.mode = mode
        self.ipv4 = ipv4
        self.ipv4_subnet = ipv4_subnet
        self.ipv6 = ipv6
        self.ipv6_prefix = ipv6_prefix
        self.host_isolated = True if host_isolated else False 
        self.isolate_guests = "yes" if isolate_guests else "no"
        self.host_mac = None
        self.ingress_route_subnet = ipaddress.ip_network(ingress_route_subnet) if ingress_route_subnet else None
        self.ingress_route_gateway = ingress_route_gateway if ingress_route_gateway else None

    def _get_config(self) -> dict:
        config = {
            "name": self.name,
            "mode": self.mode,
            "ipv4": self.ipv4,
            "ipv6": self.ipv6,
            "ipv4_subnetmask": self.ipv4_subnet,
            "ipv6_prefix": self.ipv6_prefix,
            "host_isolated": self.host_isolated,
            "isolate_guests": self.isolate_guests,
            "ingress_route_subnet_ip":  str(self.ingress_route_subnet.network_address) if self.ingress_route_subnet else None,
            "ingress_route_subnet_prefix_length": str(self.ingress_route_subnet.prefixlen) if self.ingress_route_subnet else None,
            "ingress_route_gateway": self.ingress_route_gateway 
            
        }
        return config

    def to_dict(self):
        return {"name":self.name,
                "mode":self.mode}

    def get_host_mac_from_xml(self):
        #<mac address='52:54:00:0e:4b:74'/>
        assert self.libvirt_network
        matches = re.search(r"<mac address='(?P<mac>.*)'/>", self.libvirt_network.XMLDesc())
        if not matches:
            raise Exception(f"no mac found in in network definition for {self.name} unable to set isolate host iptables rules")
        return matches.group('mac')

    def create(self, conn):
        # lookupByName throws if the name is not found
        template_path = f"{os.path.dirname(
            __file__)}/{Network.RELATIVE_TEMPLATE_PATH}"
        with open(template_path, "r") as f:
            template = jinja2.Template(f.read())
            xml_content = template.render(self._get_config())
            self.libvirt_network = conn.networkDefineXML(xml_content)
            self.libvirt_network.create()
            self.host_mac = self.get_host_mac_from_xml()
    
    def rm(self):
        assert self.libvirt_network
        if self.libvirt_network.isPersistent():
            self.libvirt_network.undefine()
        if self.libvirt_network.isActive():
            self.libvirt_network.destroy()

    @staticmethod
    def destroy_by_name(conn, name):
        network = conn.networkLookupByName(name)
        network.destroy()
        network.undefine()
        
